// feel free to change main, but be aware that we will not use your main to test your RangeAvlTree code

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.zip.*;

public class MoviePrefixSearcher {

	private static class SizeBoundedLinkedHashMap<K,V> extends LinkedHashMap<K,V> {
	     SizeBoundedLinkedHashMap(int initialCapacity, float loadFactor, boolean accessOrder) {
		super(initialCapacity,loadFactor,accessOrder);
	     }
	     protected boolean removeEldestEntry(Map.Entry eldest) {
	        return size() > CACHE_SIZE;
		}
     }

	private static final int CACHE_SIZE = 100;
	private RangeAvlTree<String> tree;
	private boolean cacheEnabled;
	private SizeBoundedLinkedHashMap<String,List<String>> cache = new SizeBoundedLinkedHashMap<String,List<String>>(134,.75f,true);

	public MoviePrefixSearcher(RangeAvlTree<String> tree) { this.tree = tree; }

	private void setCacheEnabled(boolean b) {
		cacheEnabled = b;
	}

    private static Scanner loadFromZipFile(String name) {
		ZipFile zf;
        try {
		    zf = new ZipFile(name);
		} catch (IOException e) {
			System.out.println("Couldn't find/open file?");
			return null;
		}
        ZipEntry ze = (ZipEntry)zf.entries().nextElement();
        Scanner sc = null;
        try {
        	sc = new Scanner(zf.getInputStream(ze));
		} catch (IOException e) {
			System.out.println("Couldn't create Scanner on ZipEntry?");
			return null;
		}
		return sc;
	}

	// generates the first string lexicographically after all the strings w/ prefix s
	private static String firstStringAfterPrefix(String s) {
		if (s.length()==0) throw new IllegalArgumentException("no prefix!");
		char c=s.charAt(s.length()-1);
		return s.substring(0,s.length()-1)+(char)(c+1);
	}

	// returns the number of strings in the tree starting with prefix s
	// by converting the prefix into a range
	public int numWithPrefix(String s) {
		return tree.sizeOfRange(s,firstStringAfterPrefix(s));
	}

	// returns the list of strings in order in the tree that start with prefix s
	public List<String> listWithPrefix(String s) {
		if (cacheEnabled) {
			throw new UnsupportedOperationException();
		} else {
			return tree.itemsInRange(s,firstStringAfterPrefix(s));
		}
	}

    public static void main(String[] args) {
		RangeAvlTree<String> tree=new RangeAvlTree<String>();
		MoviePrefixSearcher mps = new MoviePrefixSearcher(tree);
        Scanner sc=loadFromZipFile("moviesFiltered2.txt.zip");
        int i=0;
		while(sc.hasNextLine()) { // build the tree from the movies file
			tree.insert(sc.nextLine());
			i++;
			if (i%1000==0) System.out.println("loading data, line "+i);
		}
		System.out.println(mps.numWithPrefix("Bourne")); // prints 7
		System.out.println(mps.numWithPrefix("Batman")); // prints 32
		System.out.println(mps.numWithPrefix("Cinderella")); // prints 55
		System.out.println(mps.numWithPrefix("S")); // prints 46483

		for(String s:mps.listWithPrefix("Spartacus"))
			System.out.println(s);
			// prints Spartacus (1960)
			//          Spartacus (2003)
			//          Spartacus (2004) (TV)
			//          Spartacus: Gladiator War (2005) (TV)

        long t1=System.currentTimeMillis();
		for(int j=0;j<100;j++) {
			mps.numWithPrefix("S");
		}
        long t2=System.currentTimeMillis();
		System.out.println("time in millisec for 100 large numWithPrefix queries: "+(t2-t1));
		// this time should dramatically improve with an O(log n) implementation
		// on my laptop: O(n) version: 8072 ms; O(log n) version: <=10 ms

        t1=System.currentTimeMillis();
		for(int j=0;j<100;j++) {
			mps.listWithPrefix("Spartacus");
		}
        t2=System.currentTimeMillis();
		System.out.println("time in millisec for 100 small listWithPrefix queries: "+(t2-t1));
		// this time should dramatically improve with an O(m+log n) implementation
		// on my laptop: O(n) version: 7521 ms; O(m+log n) version: <=10 ms

		// these timing tests assume CACHE_SIZE == 100
		mps.setCacheEnabled(true);
        t1=System.currentTimeMillis();
		for(int j=0;j<1000;j++) {
			String prefix = ""+((char)('A'+j%10))+""+((char)('a'+(j/10)%10));
			mps.listWithPrefix(prefix);
		}
        t2=System.currentTimeMillis();
		System.out.println("time in millisec for 1000 queries repeating every 100 w/ caching: "+(t2-t1));
		// on my laptop with the O(n) version: 8542 ms

        t1=System.currentTimeMillis();
		for(int j=0;j<1000;j++) {
			String prefix = ""+((char)('A'+j%11))+""+((char)('a'+(j/11)%11));
			mps.listWithPrefix(prefix);
		}
        t2=System.currentTimeMillis();
		System.out.println("time in millisec for 1000 queries repeating every 121 w/ caching: "+(t2-t1));
		// on my laptop with the O(n) version: 75709 ms, since cache misses happen here
	}
}
